/*
 * Socket.C - class members for Socket class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Socket.h"
#include "ScriptContext.h"
#include "FlagRecord.h"
#include "ScriptArgument.h"
#include "Command.h"
#include "StringVariable.h"
#include "IntegerVariable.h"
#include "CoupletVariable.h"
#include "MiscVariable.h"
#define CLASS "ScriptContext"

static char * tstring;

ScriptContext::ScriptContext ( void ) : KeyedList<MISCLEN> ( ) {
   FUNCTION;
   G->variable_list = NEW ( Variable () );
   G->command_list = NEW( Command () );
   tstring = 0;
}

ScriptContext::~ScriptContext ( void ) {
   FUNCTION;
   if (!key[0]) {
      DELETE ( G->variable_list );
      DELETE ( G->command_list );
      if (tstring) {
	 my_free ( tstring, G->memtype_script );
	 tstring = 0;
      }
   } else
     G->scripts->Remove(this);
}

int ScriptContext::ExpMem ( void ) {
   FUNCTION;
   ScriptContext * ptr = this;
   Variable * var = G->variable_list;
   Command * cmd = G->command_list;
   int tot = sizeof(Variable) + sizeof(Command) + sizeof(ScriptContext)
     + (tstring ? strlen(tstring) + 1 : 0);
   
   while ((ptr = ptr->Next()))
     tot += ptr->ExpMem ();
   while ((var = var->Next()))
     tot += var->ExpMem ();
   while ((cmd = cmd->Next()))
     tot += cmd->ExpMem();
   return tot;
}
 
void ScriptContext::SetMeUp ( void ) {
   FUNCTION;
   Variable * var = G->variable_list;
   Command * cmd = G->command_list;
   
   while ((var = var->Next())) {
      if (var->Type() == Variable::STRING)
	AddStringVar((StringVariable *)var);
      else if (var->Type() == Variable::INTEGER)
	AddIntVar((IntegerVariable *)var);
      else if (var->Type() == Variable::COUPLET)
	AddCoupletVar((CoupletVariable *)var);
      else
	AddMiscVar((MiscVariable *)var);
   }
   while ((cmd = cmd->Next())) 
     AddCommand(cmd);
}

void ScriptContext::SetString ( const char * x) {
   FUNCTION;
   if (tstring)
     my_free(tstring,G->memtype_script);
   if (x) {
      tstring = (char *)my_malloc(strlen(x) + 1, G->memtype_script);
      strcpy(tstring, x);
   } else 
     tstring = 0;
}

void ScriptContext::SetInteger ( int x ) {
   FUNCTION;
   SetString(G->ToBase10(x));
}

void ScriptContext::SetArray ( int, const char **) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "SetArray"));
}

void ScriptContext::AppendString ( const char * x ) {
   FUNCTION;
   if (!tstring)
     SetString(x);
   else if (x) {
      char * ostring = tstring;
      int l1 = strlen(tstring), l = l1 + strlen(x) + 2;
	 
      tstring = (char *)my_malloc(l + 1, G->memtype_script);
      strcpy(tstring, ostring);
      tstring[l1] = ' ';
      strcpy(tstring + l1 + 1, x);
      my_free(ostring,G->memtype_script);
   }
}

void ScriptContext::AppendInteger ( int x ) {
   FUNCTION;
   AppendString(G->ToBase10(x));
}

void ScriptContext::AppendArray ( int, const char ** ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "AppendArray"));
}

void ScriptContext::AddStringVar ( StringVariable * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "AddStringVar"));
}

void ScriptContext::RemStringVar ( StringVariable * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "RemStringVar"));
}

void ScriptContext::AddIntVar ( IntegerVariable * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "AddIntVar"));
}

void ScriptContext::RemIntVar ( IntegerVariable * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "RemIntVar"));
}

void ScriptContext::AddCoupletVar ( CoupletVariable * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "AddCoupletVar"));
}

void ScriptContext::RemCoupletVar ( CoupletVariable * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "RemCoupletVar"));
}

void ScriptContext::AddMiscVar ( MiscVariable * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "AddMiscVar"));
}

void ScriptContext::RemMiscVar ( MiscVariable * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "RemMiscVar"));
}

int ScriptContext::Evaluate ( const char *, int, ... ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "Evaluate"));
   return SCRIPT_ERROR;
}

const char * ScriptContext::AsString ( void ) {
   FUNCTION;
   return safe_str(tstring);
}

int ScriptContext::AsInteger ( void ) {
   FUNCTION;
   return atoi(safe_str(tstring));
}

void ScriptContext::Clear ( void ) {
   FUNCTION;
   if (tstring) {
      my_free(tstring,G->memtype_script);
      tstring = 0;
   }
}

void ScriptContext::AddCommand ( Command * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "AddCommand"));
}

void ScriptContext::RemCommand ( Command * ) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "RemCommand"));
}

void ScriptContext::CleanupModule ( ModuleEntry * m ) {
   FUNCTION;
   Variable * optr = G->variable_list, *ptr;
   Command * ocptr = G->command_list, *cptr;
   
   if (m) {
      while ((ptr = optr->Next())) 
	if (ptr->Module() == m) {
	   DELETE (ptr);
	} else
	optr = ptr;
      while ((cptr = ocptr->Next())) 
	if (cptr->Module() == m) {
	   DELETE ( cptr );
	} else
	ocptr = cptr;
   }

}

BindReturn ScriptContext::CallBind(BindEntry *, const char *, int, va_list) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang(0x069, "s", "CallBind"));
   return NOMATCH;
}

void ScriptContext::DumpVar ( Socket * where, const char * x) {
   FUNCTION;
   Variable * var = G->variable_list->Find(x);
   char * which, buf[512];
   
   if (var) {
      if (var->Type() == Variable::STRING)
	which = ((StringVariable *)var)->buf;
      else if (var->Type() == Variable::INTEGER)
	which = G->ToBase10(*((IntegerVariable *)var)->val);
      else if (var->Type() == Variable::COUPLET) {
	 G->SPrintF(buf, 511, "%d:%d", *((CoupletVariable *)var)->lptr,
		    *((CoupletVariable *)var)->rptr);
	 which = buf;
      } else 
	which = ((MiscVariable *)var)->Read(this, 0, 0);
      where->Send(G->Lang(0x064, "s s", var->Key(), 20 - strlen(var->Key()),
		       which));
   } else 
     where->Send(G->Lang(0x0C4, "s", x));
}

void ScriptContext::DumpVars ( Socket * where ) {
   FUNCTION;
   Variable * var = G->variable_list;
   char * which, buf[512];
   
   where->Send(G->Lang(0x063, ""));
   while ((var = var->Next())) {
      if (var->Type() == Variable::STRING)
	which = ((StringVariable *)var)->buf;
      else if (var->Type() == Variable::INTEGER)
	which = G->ToBase10(*((IntegerVariable *)var)->val);
      else if (var->Type() == Variable::COUPLET) {
	 G->SPrintF(buf, 511, "%d:%d", *((CoupletVariable *)var)->lptr,
		    *((CoupletVariable *)var)->rptr);
	 which = buf;
      } else 
	which = ((MiscVariable *)var)->Read(this, 0, 0);
      where->Send(G->Lang(0x064, "s s", var->Key(), 20 - strlen(var->Key()),
		       which));
   }
   where->Send(G->Lang(0x065, ""));
}

int ScriptContext::WriteVariables ( FILE * f ) {
   FUNCTION;
   Variable * var = G->variable_list;
   int x = !EOF;
   
   while ((var = var->Next()) && (x != EOF)) {
      if (var->Type() == Variable::STRING) {
	 if (!(((StringVariable *)var)->flags & (VAR_READONLY | VAR_PROTECT))) 
	x = fprintf(f,"set %s \"%s\"\n", var->Key(),((StringVariable *)var)->buf);
      } else if (var->Type() == Variable::INTEGER) {
	 if (!(((IntegerVariable *)var)->readonly & (VAR_READONLY | VAR_PROTECT))) 
	   x = fprintf(f,"set %s %d\n", var->Key(),*((IntegerVariable *)var)->val);
      } else if (var->Type() == Variable::COUPLET) {
	 x = fprintf(f,"set %s %d:%d\n", var->Key(),*((CoupletVariable *)var)->lptr,
		     *((CoupletVariable *)var)->rptr);
      } else if (!(((MiscVariable *)var)->flags & (VAR_READONLY | VAR_PROTECT))) 
	x = fprintf(f,"set %s %s\n", var->Key(), 
		    ((MiscVariable *)var)->Read(this, 0, 0));
   }
   return x;
}
   
void ScriptContext::SetVar ( Socket * where, const char * what, const char * to) {
   FUNCTION;
   Variable * var = G->variable_list->Find(what);

   if (var) {
      if (var->Type() == Variable::STRING) {
	 StringVariable * sv = (StringVariable *)var;
	 int l = sv->length;
	 const char * list;
	 
	 if ((sv->flags & (VAR_PROTECT | VAR_READONLY))
	     && (G->protect_readonly || (sv->flags & VAR_READONLY))) {
	    where->Send(G->Lang(0x129, ""));
	    return;
	 }
	 if (*to == '"') {
	    to++;
	    list = strchr(to, '"');
	    if (!list) {
	       where->Send(G->Lang(0x0C5, ""));
	       return;
	    }
	 } else
	   for (list = to; *list && (*list != ' '); list++);
	 if (l > (list - to))
	   l = list - to;
	 strncpyz(sv->buf, to, l);
	 if ((sv->flags & VAR_DIR) && to[0]) {
	    l--;
	    if ((l > 0) && (sv->buf[l] != '/')) {
	       sv->buf[++l] = '/';
	       sv->buf[++l] = 0;
	    }
	 }
      } else if (var->Type() == Variable::INTEGER) {
	 IntegerVariable * iv = (IntegerVariable *)var;
	 if (iv->val == (int *)&G->conmask) {
	    if (to[0])
	      G->conmask.Set(to);
	    else
	      G->conmask.Set ( LogModes::MODES | LogModes::MISC | LogModes::CMDS);
	 } else if (iv->val == &G->default_flags) {
	    FlagRecord fr;
	    fr.BreakDown(to);
	    G->default_flags = fr.Global();
	 } else if ((iv->readonly == VAR_READONLY) 
		    || ((iv->readonly == VAR_PROTECT) && G->protect_readonly))
	     where->Send(G->Lang(0x129, ""));
	 else 
	   *(iv->val) = atoi(to);
      } else if (var->Type() == Variable::COUPLET) {
	 const char * p = strchr(to, ':');
	 CoupletVariable * cp = (CoupletVariable *)var;
	 if (p) {
	    p++;
	    *(cp->lptr) = atoi(p);
	    *(cp->rptr) = atoi(to);
	 } else
	   *(cp->lptr) = atoi(to);
      } else {
	 MiscVariable * mv = (MiscVariable *)var;
	 char * res = 0;
	 
	 if (mv->flags & VAR_LIST) {
	    int lc = 0;
	    const char * list;
	    char buf[512];
	    ScriptArgument * args[255];
	    
	    while ((lc < 255) && to && *to) {
	       while (*to == ' ')
		 to++;
	       if (*to == '"') {
		  to++;
		  list = strchr(to, '"');
		  if (!list) {
		     where->Send(G->Lang(0x0C5, ""));
		     break;
		  }
		  strncpyz(buf, to, list - to );
		  args[lc] = NEW ( ScriptArgument ( this, buf, 1 ) );
		  to = list + 1;
	       } else {
		  for (list = to; *list && (*list != ' '); list++);
		  strncpyz(buf, to, list - to);
		  args[lc] = NEW ( ScriptArgument ( this, buf, 1 ) );
		  to = list;
	       }
	       lc++;
	    }
	    args[lc] = 0;
	    if ((res = mv->Set(this, lc, args)))
	      where->Send(res);
	    while (lc) {
	       lc--;
	       DELETE ( args[lc] );
	    }
	 } else {
	    ScriptArgument pass(this, to), * args[2] = {&pass, 0};
	    char buf[512];
	    
	    if (*to == '"') {
	       const char * list;
	       
	       to++;
	       list = strchr(to, '"');
	       if (!list) {
		  where->Send(G->Lang(0x0C5, ""));
		  return;
	       }
	       strncpyz(buf, to, list - to);
	       pass.ptr = buf;
	    }
	    if ((res = mv->Set(this, 1, args)))  
	      where->Send(res);
	 }
      }
   } else
     where->Send(G->Lang(0x0C4, "s", what));
}

int ScriptContext::ParseFile ( const char * fname) {
   FUNCTION;
   char buf[512], *ptr, *cmd, *what;
   Command * c;
   int line = 0, i;
   FILE * f;
   
   if (!(f = fopen(fname, "r")))
     return SCRIPT_ERROR;
   fgets(buf, 512, f);
   while (!feof(f)) {
      if (buf[(i = strlen(buf) - 1)] == '\n')
	buf[i] = 0;
      line ++;
      for (ptr = buf; *ptr == ' '; ptr++);
      if (*ptr && (*ptr != '#')) {
	 cmd = G->Split(&ptr);
	 if (!strcasecmp(cmd, "set")) {
	    what = G->Split(&ptr);
	    SetVar(G->Log, what, ptr);
	 } else if (!strcasecmp(cmd, "source")) {
	    what = G->Split(&ptr);
	    if (ParseFile(what) == SCRIPT_ERROR)
	      return SCRIPT_ERROR;
	 } else {
	    if ((c = G->command_list->Find(cmd))) {
	       int lc = 1;
	       char * list;
	       ScriptArgument * args[255];
	       
	       args[0] = NEW ( ScriptArgument ( this, cmd) );
	       while ((lc < 255) && ptr && *ptr) {
		  while (*ptr == ' ')
		    ptr++;
		  if (*ptr == '"') {
		     ptr++;
		     list = strchr(ptr, '"');
		     if (!list) {
			G->Log->Send(G->Lang(0x0C5, ""));
			break;
		     }
		     *list = 0;
		     args[lc] = NEW ( ScriptArgument ( this, ptr ) );
		     ptr = list + 1;
		  } else {
		     list = G->Split(&ptr);
		     args[lc] = NEW ( ScriptArgument ( this, list ) );
		  }
		  lc++;
	       }
	       args[lc] = 0;
	       i = c->func(this, lc, args);
	       while (lc--) {
		 DELETE ( args[lc] );
	       }
	       if (i == SCRIPT_ERROR)
		 return SCRIPT_ERROR;
	    } else 
	      G->Fatal(G->Lang(0x0C7, "sssd", cmd, ptr, fname, line));
	 }
      }
      fgets(buf, 512, f);
   }
   return SCRIPT_OK;
}
