/*
 * Socket.C - class members for Socket class
 */
/*
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Socket.h"
#include "UserRecord.h"
#include "HostnameCache.h"
#include "FlagRecord.h"
#include "LogFile.h"
#include <sys/socket.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <errno.h>
#if HAVE_GETRUSAGE
#include <sys/resource.h>
#if HAVE_SYS_RUSAGE_H
#include <sys/rusage.h>
#endif
#endif
#ifdef HAVE_UNAME
#include <sys/utsname.h>
#endif
#define CLASS "Socket"

const char * Socket::SOCKET_NONSOCK = "NON_SOCKET";

Socket::Socket ( Socket & old, const char * ntype, int tflags, time_t * tv )
: IntList ( *G->socklist, old, IntList::APPEND ) {
   FUNCTION;
   type = ntype;
   flags = old.flags;
   type_flags = tflags;
   status = old.status;
   address = old.address;
   localport = old.localport;
   remoteport = old.remoteport;
   user = old.user;
   strcpy(nick, old.nick);
   strcpy(host, old.host);
   timeval = G->now;
   timeout_val = tv;
   inbuflen = old.inbuflen;
   outbuflen = old.outbuflen;
   outbufoff = old.outbufoff;
   outbufsize = old.outbufsize;
   outbuf = old.outbuf;
   old.outbuf = 0;
   memcpy(inbuf, old.inbuf, 514);
   old.key = 0;
   old.flags |= NONSOCK;
   old.Close();
}

Socket::Socket ( int fgs, const char * ntype, int tflags, time_t * tv,
		int fd, const char * nhost, int rpt)
: IntList ( *G->socklist, IntList::APPEND ) {
   FUNCTION;
   if (Init (fgs, fd)) {
      type = ntype;
      type_flags = tflags;
      timeout_val = tv;
      if (nhost) {
	 strncpy(host, nhost, UHOSTLEN);
	 host[UHOSTLEN] = 0;
      }
      remoteport = rpt;
   } else 
     G->Fatal(G->Lang( 0x001, ""));
}

Socket::Socket ( const char * ntype, int flags, struct in_addr addr, int port,
		int tflags, time_t * tv)
: IntList ( *G->socklist, IntList::APPEND ) {
   FUNCTION;
   struct sockaddr_in sain;
#if __GLIBC__ >= 2
   unsigned
#endif
     int addrlen = sizeof(struct sockaddr_in);
   const char * name;
   struct in_addr * in_ptr;
   
   if (Init (Socket::CONNECTING | flags, 0 ) ) {
      type = ntype;
      type_flags = tflags;
      timeout_val = tv;
      memset((char *) &sain, 0, sizeof(struct sockaddr_in));
      sain.sin_family = AF_INET;
      sain.sin_port = 0;
      sain.sin_addr.s_addr = G->GetMyIp()->s_addr;
      if ((bind(key, (struct sockaddr *)&sain,sizeof(struct sockaddr_in)) >= 0)
	  && (getsockname(key, (struct sockaddr *) &sain, &addrlen) >= 0)) {
	 localport = htons(sain.sin_port);
	 if (G->firewall[0]) {
	    in_ptr = G->host_cache->Find(G->firewall + 
					 (G->firewall[0] == '!' ? 1 : 0));
	    if (in_ptr) {
	       sain.sin_addr.s_addr = in_ptr->s_addr;
	       sain.sin_port = htons(G->firewallport);
	    } else {
	       Close();
	       return;
	    }
	 } else {
	    sain.sin_port = htons(port);
	    sain.sin_addr.s_addr = htonl(addr.s_addr);
	 }
	 if ((connect(key, (struct sockaddr *)&sain,
		     sizeof(struct sockaddr_in)) >= 0)
	     || (errno == EINPROGRESS)) {
	    address = addr;
	    if ((name = G->host_cache->Find(&sain.sin_addr))) {
	       strncpy(host, name, UHOSTLEN);
	       host[UHOSTLEN] = 0;
	    } else
	      strcpy(host, "*");
	    remoteport = port;
	    G->LOG_SOCKET.Put(G->Lang( 0x002, "Xddd", address.s_addr,
				      remoteport, localport, key));
	    if (G->firewall[0]) 
	      SendFirewall(&addr, port);
	 } else 
	   Close();
	 return;
      }
   }
   G->Fatal(G->Lang( 0x001, ""));
}

Socket::~Socket ( void ) {
   FUNCTION;
   G->socklist->Remove ( this );
   next = 0;
   if ((key >= 0) && !(flags & Socket::NONSOCK)) {
      close(key);
      key = -1;
   }
   G->dcc_total --;
}

int Socket::Init ( int fgs, int x) {
   FUNCTION;
   int parm = 1;

   if (G->dcc_total < G->max_dcc) {
      if (fgs & (NONSOCK | FD_)) {
	 key = x;
	 if (fgs & FD_)
	   FD_SET(key,&G->fds_read);
      } else if ((key = socket(AF_INET,SOCK_STREAM,0)) >= 0)
	FD_SET(key,&G->fds_read);
      if (!(fgs & NONSOCK) && (key || G->backgrd)) {
	 setsockopt(key, SOL_SOCKET, SO_KEEPALIVE, (void *) &parm, sizeof(int));
	 parm = 0;
	 setsockopt(key, SOL_SOCKET, SO_LINGER, (void *) &parm, sizeof(int));
      }
      if ((fgs & NONSOCK) || (fcntl(key,F_SETFL,O_NONBLOCK) >= 0)) {
	 flags = fgs;
	 G->dcc_total ++;
	 return 1;
      }
   }
   return 0;
}

void Socket::Close ( void ) {
   FUNCTION;
   if (!(flags & Socket::NONSOCK)) {
      FD_CLR(key,&G->fds_read);
      FD_CLR(key,&G->fds_write);
   }
   flags |= LOST;
   type = 0;
   type_flags = 0;
   if (outbuf) {
      my_free(outbuf,G->memtype_socket);
      outbuf = 0;
   }
   if ((key >= 0) && !(flags & Socket::NONSOCK)) {
      close(key);
      key = -1;
   }
}

int Socket::Scan ( int max ) {
   FUNCTION;
   struct sockaddr_in sain, sain2;
   struct in_addr * in_ptr, *in_pt2;
#if __GLIBC__ >= 2
   unsigned
#endif
     int addrlen = sizeof(struct sockaddr_in), parm = 1;
   LogModes Socklog ( LogModes::SOCKET );
   
   memset((char *) &sain, 0, sizeof(struct sockaddr_in));
   memset((char *) &sain2, 0, sizeof(struct sockaddr_in));
   sain.sin_family = AF_INET;
   sain2.sin_family = AF_INET;
   if ((in_pt2 = G->host_cache->Find(host))) {
      address.s_addr = htonl(in_pt2->s_addr);
      while (remoteport < max) {
	 if (G->firewall[0]) {
	    in_ptr = G->host_cache->Find(G->firewall + 
					 (G->firewall[0] == '!' ? 1 : 0));
	    sain.sin_port = htons(G->firewallport);
	 } else {
	    in_ptr = in_pt2;
	    sain.sin_port = htons(remoteport);
	 }
	 if (!in_ptr)
	   return 1;
	 sain.sin_addr.s_addr = in_ptr->s_addr;
	 if (key) {
	    FD_CLR(key,&G->fds_read);
	    FD_CLR(key,&G->fds_write);
	    close (key);
	 }
	 if ((key = socket(AF_INET,SOCK_STREAM,0)) < 0)
	   return 1;
	 FD_SET(key,&G->fds_read);
	 setsockopt(key, SOL_SOCKET, SO_KEEPALIVE, (void *) &parm, sizeof(int));
	 parm = 0;
	 setsockopt(key, SOL_SOCKET, SO_LINGER, (void *) &parm, sizeof(int));
	 if (fcntl(key, F_SETFL, O_NONBLOCK) < 0)
	   return 1;
	 sain2.sin_port = 0;
	 sain2.sin_addr.s_addr = G->GetMyIp()->s_addr;
	 if ((bind(key, (struct sockaddr *)&sain2,
		   sizeof(struct sockaddr_in)) < 0) || 
	     (getsockname(key, (struct sockaddr *) &sain2, &addrlen) < 0))
	   return 1;
	 localport = htons(sain2.sin_port);
	 Socklog.Put(G->Lang( 0x002, "Xddd", address.s_addr,
			     remoteport, localport, key));
	 if ((connect(key, (struct sockaddr *)&sain,
		      sizeof(struct sockaddr_in)) >= 0)
	     || (errno == EINPROGRESS)) {
	    if (G->firewall[0]) 
	      SendFirewall(in_pt2, remoteport);
	    return 0;
	 }
	 remoteport++;
      }
   } else
     return 2;
   return 1;
}

int Socket::Write( char * buf, int inlen ) {
   int len;
   LogModes Socklog ( LogModes::SOCKET );
 
   if (flags & NONSOCK) 
     return write(key, buf, inlen);
   if (flags & LOST)
     return -1;
   if ((key == STDIN) && !G->backgrd)
     return write(STDOUT, buf, inlen);
   if (inlen == -1)
     inlen = strlen(buf);
   len = inlen;
   if (!outbuflen) {
      int written = write(key, buf, len);
      
      if ((written < 0) && (errno != EAGAIN)
#ifdef EBADSLT 
	  && (errno != EBADSLT)
#endif
#ifdef ENOTCONN
	  && (errno != ENOTCONN)
#endif
	  ) {
	 /* error of some sort */
	 /* clear it forreading, therefore preventing recursize errors */
	 FD_CLR(key,&G->fds_read);
	 Eof();
	 Socklog.Put(G->Lang( 0x004, "Xddd", address.s_addr,
			  remoteport, localport, key));
	 return -1;
      } else if (written > 0) {
	 buf += written;
	 len -= written;
      }
   }
   if (len) {
      FD_SET(key,&G->fds_write);
      if (!outbuf) {
	 /* no buffer at all, lets make one */
	 outbufsize = ((len / 512) + 1) * 512;
	 outbuf = (char *)my_malloc(outbufsize,G->memtype_socket);
	 outbufoff = 0;
	 outbuflen = len;
	 memcpy(outbuf,buf,len);	 
      } else if ((outbuflen + outbufoff + len) < outbufsize) {
	 /* easy, just append in the current buffer */
	 memcpy(outbuf + outbufoff + outbuflen, buf, len);
	 outbuflen += len;
      } else if ((outbuflen + len) < outbufsize) {
	 /* big enough to fit in buffer, squish up buffer */
	 memcpy(outbuf, outbuf+outbufoff,outbuflen);
	 outbufoff = 0;
	 memcpy(outbuf+outbuflen,buf, len);
	 outbuflen += len;
      } else {
	 char * newbuf;
	 /* there is a buffer, but it's not big enough, expand, with
	  * a little extra ... say 512 bytes >:) */
	 while ((outbuflen + len) > outbufsize)
	   outbufsize += 512;
	 newbuf = (char *)my_malloc(outbufsize,G->memtype_socket);
	 memcpy(newbuf, outbuf + outbufoff, outbuflen);
	 my_free(outbuf,G->memtype_socket);
	 outbuf = newbuf;
	 outbufoff = 0;
	 memcpy(outbuf+outbuflen, buf, len);
	 outbuflen += len;
      }
   }
   return inlen;
}

void Socket::Send ( char * buf ) {
   FUNCTION;
   int len = strlen(buf);
   
   if (!(flags & LOST)) {
      if (len > 500)
	len = 500;
      if (buf[len - 1] != '\n') {
	 buf[len++] = '\n';
	 buf[len] = 0;
      }
      Dump (buf, len);
   }
}

void Socket::Printf ( char * format, ... ) {
   /* FUNCTION; - Recursive call, dont use */
#ifdef HAVE_VSNPRINTF
   char buffer[512];
#else
   char buffer[1024]; /* we have to hope this is enough */
#endif
   int len;
   va_list valist;
    
   if (!(flags & LOST)) {
      va_start(valist, format);
#ifdef HAVE_VSNPRINTF
      if ((len = vsnprintf(buffer, 511, format, valist)) < 0)
	buffer[len = 511] = 0;
#else
      if ((len = vsprintf(buffer, format, valist)) > 511)
	buffer[len = 511] = 0;
      
#endif
      va_end(valist);
      Dump(buffer, len);
   }
}
   
int Socket::ExpMem ( void ) {
   FUNCTION;
   Socket * ptr = this;
   int tot = sizeof(Socket);
   
   if (ptr == G->socklist) 
     while ((ptr = ptr->Next()))
       tot += ptr->ExpMem() + ptr->outbufsize;
   return tot;
}

void Socket::Display ( Socket * where ) {
   FUNCTION;
   Socket * ptr = this;
    
   where->Send(G->Lang( 0x005, ""));
   where->Send(G->Lang( 0x006, ""));
   while ((ptr = ptr->Next())) {
      char bufflag[9], thost[19];
      
      if (ptr->flags & Socket::NONSOCK)
	strcpy(bufflag,"-NOSOCK-");
      else {
	 bufflag[0] = (ptr->flags & CONNECTING) ? 'C':'-';
	 bufflag[1] = (ptr->flags & LISTENING) ? 'L':'-';
	 bufflag[2] = (ptr->flags & STRONG) ? 'S':'-';
	 bufflag[3] = (ptr->flags & LOST) ? 'L':'-';
	 bufflag[4] = (ptr->flags & RAW) ? 'R':'-';
	 bufflag[5] = (ptr->flags & STALE) ? 'S':'-';
	 bufflag[6] = FD_ISSET(ptr->key,&G->fds_read) ? 'R' : '-';
	 bufflag[7] = FD_ISSET(ptr->key,&G->fds_write) ? 'W' : '-';
	 bufflag[8] = 0;
      }
      strncpy(thost, ptr->host[0] ? ptr->host : "*", 18);
      thost[18] = 0;
      where->Send(G->Lang( 0x007, "255Xs ss", ptr->key, ptr->localport,
		       ptr->remoteport, ptr->address.s_addr, thost, 
		       18 - strlen(thost), bufflag,
		       G->Lang( 0x008, "5555", ptr->inbuflen, ptr->outbuflen, 
			    ptr->outbufsize, ptr->outbufoff)));
      ptr->Display2(where);
   }
}

Socket * Socket::Find ( int fd, const char * type ) {
   FUNCTION;
   Socket * ptr = this;
   
   while ((ptr = ptr->Next()))
     if ((fd == ptr->FD()) && (type == ptr->Type()))
       break;
   return ptr;
}

Socket * Socket::Find ( const char * str, int flgs ) {
   FUNCTION;
   Socket * ptr = this;
   
   while ((ptr = ptr->Next()))
     if (!strcasecmp(str, ptr->Nick()) && (!flgs || (ptr->TypeFlags() & flgs)))
       break;
   return ptr;
}


/* dump a potentially super-long string of text */
/* assume prefix 20 chars or less */
void Socket::DumpLots (char * prefix, char * data) {
   FUNCTION;
   char *n, c, * p = data, * q;
   
   if (!(*data)) 
      Printf("%s\n", prefix);
   else {
      while (strlen(p) > 480) {
	 q = p + 480;
	 /* search for embedded linefeed first */
	 n = strchr(p, '\n');
	 if (n && (n < q)) {
	    /* great! dump that first line then start over */
	    *n = 0;
	    Printf("%s%s\n", prefix, p);
	    *n = '\n';
	    p = n + 1;
	 } else {
	    /* search backwards for the last space */
	    while ((*q != ' ') && (q != p))
	      q--;
	    if (q == p)
	      q = p + 480;
	    /* ^ 1 char will get squashed cos there was no space -- too bad */
	    c = *q;
	    *q = 0;
	    Printf("%s%s\n", prefix, p);
	    *q = c;
	    p = q + 1;
	 }
      }
      /* last trailing bit: split by linefeeds if possible */
      while ((n = strchr(p, '\n'))) {
	 *n = 0;
	 Printf("%s%s\n", prefix, p);
	 *n = '\n';
	 p = n + 1;
      }
      if (*p)
	Printf("%s%s\n", prefix, p);	/* last trailing bit */
   }
}

/* get icon symbol for a user (depending on access level) */
/* (*)owner on any channel  (+)master on any channel (%) botnet master */
/* (@)op on any channel  (-)other  */
char Socket::GetIcon ( void ) {
   FUNCTION;
   FlagRecord fr ( FlagRecord::GLOBAL | FlagRecord::CHANNEL
		  | FlagRecord::ANYWHERE);
   char x = '-';
   
   if (user) {
      fr.Set(user);
      
      if (fr.GlobOwner())
	x = 'N';
      else if (fr.ChanOwner())
	x = 'n';
      else if (fr.GlobMaster())
	x = 'M';
      else if (fr.ChanMaster())
	x = 'm';
      else if (fr.GlobBotmast())
	x = 'T';
      else if (fr.GlobOp())
	x = 'O';
      else if (fr.ChanOp())
	x = 'o';
   }
   return x;
}


void Socket::Eof (void) {
   FUNCTION;
   Close ();
}

void Socket::Activity (char *, int) {
   FUNCTION;
   G->LOG_SOCKET.Put(G->Lang( 0x00C, "Xddd", address.s_addr, remoteport,
			   localport, Key()));
}

void Socket::Display2 (Socket * where) {
   FUNCTION;
   where->Send(G->Lang( 0x009, "s", type));
}

void Socket::Timeout (void) {
   FUNCTION;
   Eof();
}

int Socket::Dump ( char * buf, int len ) {
   return Write(buf, len);
}

void Socket::StripTelnet (char * buf, int * len) {
   FUNCTION;
   unsigned char *p = (unsigned char *) buf, *o = (unsigned char *)buf;
   int mark;
   
   while (*p != 0) {
      while ((*p != 255) && (*p != 0))
	 *o++ = *p++;
      if (*p == 255) {
	 p++;
	 mark = 2;
	 if (!*p)
	    mark = 1;		/* bogus */
	 else if ((*p >= 251) && (*p <= 254)) {
	    mark = 3;
	    if (!*(p + 1))
	       mark = 2;	/* bogus */
	 } else if ((*p == 251) && !(*(p + 1) == 1)) {
	    /* WILL X -> response: DONT X */
	    /* except WILL ECHO which we just smile and ignore */
	    write(key, "\377\376", 2);
	    write(key, p + 1, 1);
	 } else if ((*p == 253) && !(*(p + 1) == 1)) {
	    /* DO X -> response: WONT X */
	    /* except DO ECHO which we just smile and ignore */
	    write(key, "\377\374", 2);
	    write(key, p + 1, 1);
	 } else if (*p == 246) 
	   /* "are you there?" */
	   /* response is: "hell yes!" */
	   write(key, "\r\nHell, yes!\r\n", 14);
	 /* anything else can probably be ignored */
	 p += mark-1;
	 *len = *len - mark;
      }
   }
   *o = *p;
}

int Socket::Handle ( void ) {
   FUNCTION;
   fd_set fdt_read, fdt_write;
   struct timeval tv;
   Socket * sock = G->socklist, * sock2 = sock->Next();
   static int highfd = 0, redo = 0;
   static char null_string[1] = "";
   int check_all;
   
   if (!(check_all = redo)) {
      fdt_read = G->fds_read;
      fdt_write = G->fds_write;
      tv.tv_sec = 1;
      tv.tv_usec = 0;
#ifdef HPUX_HACKS
# ifndef HPUX10_HACKS
      if (select(highfd + 1, (int *) &fdt_read, (int *) &fdt_write, NULL, &tv) < 0)
# else
      if (select(highfd + 1, &fdt_read, &fdt_write, NULL, &tv) < 0)
# endif
#else
      if (select(highfd + 1, &fdt_read, &fdt_write, NULL, &tv) < 0)
#endif
	return -1;
   }
   highfd = 0;
   redo = 0;
   while ((sock = sock2)) {
      sock2 = sock->Next();
      if (sock->flags & LOST) {
	 DELETE( sock );
	 continue;
      } else if (sock->flags & NONSOCK) 
	   continue;
      if (sock->Key() > highfd)
	highfd = sock->Key();
      if (!check_all) {
	 if (FD_ISSET(sock->Key(),&fdt_write)) {
	    int len = 512;
	    if (sock->outbuflen) {
	       if (len > sock->outbuflen)
		 len = sock->outbuflen;
	       if (((len = write(sock->Key(),sock->outbuf + sock->outbufoff,len))
		    < 0) && (errno != EAGAIN) && (errno != EINPROGRESS)
#ifdef EBADSLT 
		   && (errno != EBADSLT)
#endif
#ifdef ENOTCONN
		   && (errno != ENOTCONN)
#endif
		   ) {
		  G->LOG_SOCKET.Put(G->Lang( 0x00A, "Xddd",
					  sock->address.s_addr, sock->remoteport,
					  sock->localport, sock->Key()));
		  sock->Eof();
		  continue;
	       } else if (len > 0) {
		  sock->outbufoff += len;
		  sock->outbuflen -= len;
	       }
	    }
	    if (!sock->outbuflen) { 
	       FD_CLR(sock->Key(),&G->fds_write);
	       sock->outbufoff = 0;
	       my_free(sock->outbuf,G->memtype_socket);
	       sock->outbuf = 0;
	       sock->outbufsize = 0;
	    }
	 }
	 if (FD_ISSET(sock->Key(),&fdt_read)) {
	    int len = 511 - sock->inbuflen;
	    if (sock->flags & LISTENING) 
	       sock->Activity(sock->inbuf, sock->inbuflen);
	    else if (len) {
	       if (((len = read(sock->Key(), sock->inbuf + sock->inbuflen, len)) <= 0)
		   && (errno != EAGAIN)) {
		  G->LOG_SOCKET.Put(G->Lang( 0x00B, "Xddd",	
					  sock->address.s_addr, sock->remoteport,
					  sock->localport, sock->Key()));
		  sock->Eof();
		  continue;
	       } else if (len > 0) {
		  sock->inbuflen += len;
		  sock->flags &= ~STALE;
	       }
	    }
	 }
      }
      if (!(sock->flags & (STALE | LOST)) && sock->inbuflen) {
	 if ((sock->flags & (CONNECTING | STRONG))
	     == (CONNECTING | STRONG))
	   sock->Activity(null_string, 0);
	 if ((sock->flags & CONNECTING) && G->firewall[0] && 
	     (G->firewall[0] != '!')) {
	    if (sock->inbuf[1] != 90) {
	       errno = (sock->inbuf[1] == 91) ? ECONNREFUSED : ENETUNREACH;
	       G->LOG_SOCKET.Put(G->Lang( 0x00D, "Xddd",
					 sock->address.s_addr, sock->remoteport,
					 sock->localport, sock->Key()));
	       sock->Eof();
	       continue;
	    }
	    memcpy(sock->inbuf,sock->inbuf + 8, sock->inbuflen -= 8);
	 }
	 if (sock->flags & CONNECTING)
	   sock->flags &= ~CONNECTING;
	 if (sock->flags & RAW) {
	    sock->Activity(sock->inbuf, sock->inbuflen);
	    sock->inbuflen = 0;
	 } else {
	    char * p, tempbuf[512];
	    sock->inbuf[sock->inbuflen] = 0;
	    for (p = sock->inbuf; *p && (*p != '\n') && (*p != '\r'); p++);
	    if (*p || (sock->inbuflen == 511)) {
	       int l = p - sock->inbuf, ol = l;
	       
	       memcpy(tempbuf,sock->inbuf,l);
	       tempbuf[l] = 0;
	       if (((*p == '\n') && (p[1] == '\r')) || 
		   ((*p == '\r') && (p[1] == '\n'))) {
		  p++;
		  l++;
	       }
	       if ((sock->inbuflen -= l + 1) > 0) {
		  memcpy(sock->inbuf,sock->inbuf+l+1,sock->inbuflen);
		  redo = 1;
	       } else 
		 sock->inbuflen = 0;
	       sock->Activity(tempbuf, ol);
	    } else
	      sock->flags |= STALE;
	    return 0;
	 }
      }
   }
   return 0;
}

/* dump status info out to dcc */
void Socket::TellVerboseStatus (int all) {
   FUNCTION;
   char s[256], s2[81];
   char * vers_t, * uni_t;
   int i, hr, min, s1;
   time_t now2;
#if HAVE_GETRUSAGE
   struct rusage ru;
#else
#if HAVE_CLOCK
   clock_t cl;
#endif
#endif
#ifdef HAVE_UNAME
   struct utsname un;
   FlagRecord fr;
   
   if (!uname(&un) < 0) {
#endif
      vers_t = " ";
      uni_t = "*unknown*";
#ifdef HAVE_UNAME
   } else {
      vers_t = un.release;
      uni_t = un.sysname;
   }
#endif

   i = G->userlist->Count();
   Send(G->Lang(0x1F5, "ssdd", G->origbotname[0] ? G->origbotname : G->botnetnick, 
		G->VersionStr(), i, (int) (G->memtype_list->ExpectedMemory() / 1024)));
   Send(G->Lang(0x1F6, "ss", uni_t, vers_t));
   if (G->admin[0])
     Send(G->Lang(0x1F7, "s", G->admin));
   now2 = G->now - G->online_since;
   s[0] = 0;
   i = 0;
   if (now2 > 86400) {
      i += G->SPrintF(s, 255, "%s, ", G->Lang(((now2 / 86400) >= 2) ? 0x1F8 
					      : 0x1F9, "l", (now2 / 86400)));
      now2 = now2 % 86400;
   }
   hr = now2 / 3600;
   min = (now2  % 3600) / 60;
   G->SPrintF(s + i, 255 - i, "%2:%2", hr, min);
   s1 = (G->backgrd ? 0x1FA : (G->term_z ? 0x1FB : (G->con_chan ? 0x1FC : 0x1FD)));
#if HAVE_GETRUSAGE
   getrusage(RUSAGE_SELF, &ru);
   hr = (ru.ru_utime.tv_sec + ru.ru_stime.tv_sec) / 60;
   min = (ru.ru_utime.tv_sec + ru.ru_stime.tv_sec) % 60;
   G->SPrintF(s2, 80, "CPU %2:%2", hr, min);	/* actually min/sec */
#else
#if HAVE_CLOCK
   cl = (clock() / CLOCKS_PER_SEC);
   hr = (int) cl / 60;
#ifdef HPUX_HACKS
   min = (int) (cl - (hr * 60));
#else
   min  cl % 60;
#endif
   G->SPrintF(s2, 80, "CPU %2:%2", hr, min);	/* actually min/sec */
#else
   strcpy(s2, "CPU ???");
#endif
#endif
   Send(G->Lang(0x1FE, "sssF", s, G->Lang( s1, ""), s2, 100.0 * 
		((float) G->cache_hit) / ((float) (G->cache_hit + G->cache_miss))));
   G->memtype_list->Status(this);
   if (all) {
      Printf("\n");
      Send(G->Lang(0x200, "s", G->botnetnick));
      if (G->firewall[0])
	Send(G->Lang(0x209, "sd", G->firewall, G->firewallport));
      Send(G->Lang(0x20A, "ss", G->userfile, G->motdfile));
      Send(G->Lang(0x20B, ""));
      Send(G->Lang(0x20C, "s", G->helpdir));
      Send(G->Lang(0x20D, "s", G->tempdir));
#ifndef STATIC
      Send(G->Lang(0x20E, "s", G->moddir));
#endif
      fr.SetGlobal(G->default_flags);
      fr.Build(s);
      Send(G->Lang(0x20F, "ss", s, G->notify_new));
      if (G->owner[0])
	Send(G->Lang(0x210, "s", G->owner));
      G->log_list->Display(this);
      Send(G->Lang(0x212, "d", G->ignore_time));
   } 
   G->module_list->Report(this,all,NULL);
}

void Socket::SendFirewall ( struct in_addr * where, int port ) {
   FUNCTION;
   char s[128];
   unsigned char x[10];
   
   if (G->firewall[0] == '!') {
      sprintf(s, "%s %d\n", inet_ntoa(*where), port);
      write(key, s, strlen(s));
   } else if (G->firewall[0]) {
      G->MemCpy((char *)x, (char *) where, sizeof(struct in_addr));
      sprintf(s, "\004\001%c%c%c%c%c%c%s", (port >> 8) % 256, (port % 256),
	      x[0], x[1], x[2], x[3], G->botuser);
      write(key, s, strlen(s) + 1);
   }
}
