/*
 * TelnetPort.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "TelnetPort.h"
#include "IdentWait.h"
#include "IdentLookup.h"
#include "IgnoreRecord.h"
#include "HostnameCache.h"
#include "ScriptContext.h"
#include <errno.h>
#define CLASS "TelnetPort"

static in_addr t = { INADDR_ANY };

TelnetPort::TelnetPort ( ScriptContext * s, const char * host, int port, 
			const char * mask, const char * type ) 
:  ListeningSocket ( G->TELNET_PORT, (host[0] == '*') ? &t :
		    G->host_cache->Find(host), port) {
   FUNCTION;
   sc = s;
   for (origport = port; localport < (origport + G->port_range); localport++)
     if (Enable())  {
	SetNick(type);
	SetHost(mask ? mask : "*");
	G->LOG_MISC.Put(G->Lang( 0x1AB, "Xds", address.s_addr, localport, nick));
	return;
     }
   sc->SetString(G->Lang( 0x1AA, "Xd", address.s_addr, localport));
}

TelnetPort::~TelnetPort ( void ) {
}

/* a modified detect_flood for incoming telnet flood protection */
int TelnetPort::DetectFlood (const char * floodhost) {
   FUNCTION;
   static int lasttelnets;
   static char lasttelnethost[81];
   static time_t lasttelnettime;
   
   context;
   if (G->flood_telnet_thr) {
      if (strcasecmp(lasttelnethost, floodhost)) {	/* new */
	 strcpy(lasttelnethost, floodhost);
	 lasttelnettime = G->now;
	 lasttelnets = 0;
      } else if (lasttelnettime < G->now - G->flood_telnet_time) {
	 /* flood timer expired, reset it */
	 lasttelnettime = G->now;
	 lasttelnets = 0;
      } else if (++lasttelnets >= G->flood_telnet_thr) {	/* FLOOD */
	 /* reset counters */
	 lasttelnets = 0;
	 lasttelnettime = 0;
	 lasttelnethost[0] = 0;
	 G->LOG_MISC.Put(G->G->Lang( 0x28E, "s", floodhost));
	 NEW ( IgnoreRecord ( floodhost, G->origbotname,
			     "Telnet connection flood", 
			     G->now + (60 * G->ignore_time)) );
	 return 1;
      }
   }
   return 0;
}

Socket * TelnetPort::Create ( int fd ) {
   FUNCTION;
   return NEW ( IdentWait ( fd, this ) );
}

void TelnetPort::Activity (char *, int) {
   FUNCTION;
   char s1[UHOSTLEN + 1], *s = 0;
   IdentWait * ptr;
   IdentLookup * ident;
   
   context;
   if (G->dcc_total >= G->max_dcc) {
      G->dcc_total--;
      while ((ptr = (IdentWait *)Answer())) {
	 ptr->Send(G->Lang( 0x15C, ""));
	 ptr->Close();
      }
      G->dcc_total++;
   } else {
      context;
      if (!(ptr = (IdentWait *)Answer()) && (errno == EAGAIN))
	return;
      else if (!ptr)
	G->LOG_MISC.Put(G->Lang( 0x15D, "s", strerror(errno)));
      else if ((host[0] == '@') && !G->WildMatch(host + 1, ptr->Host())) {
	 G->LOG_MISC.Put(G->Lang( 0x15E, "s", ptr->Host()));
	 ptr->Close();
      } else {
	 context;
	 G->SPrintF(s1, UHOSTLEN, "telnet!telnet@%s",ptr->Host());
	 if (G->ignores->Match(s1) || DetectFlood(s1)) 
	   ptr->Close();
	 else {
	    context;
	    G->LOG_MISC.Put(G->Lang( 0x15F, "sd", ptr->Host(), ptr->RemotePort()));
	    if (G->dcc_total >= G->max_dcc)
	      s = G->Lang( 0x160, "");
	    ident = NEW ( IdentLookup ( *ptr ) );
	    if (ident->Flags() & Socket::LOST)
	      s = strerror(errno);
	    else {
	       ident->Printf("%d, %d\n",ptr->RemotePort(),localport);
	       FD_CLR(ptr->FD(), &G->fds_read);
	    }
	    if (s) {
	       G->LOG_MISC.Put(G->Lang( 0x161, "ss", ptr->Host(), s));
	       ptr->GotResult(s1+7);
	    }
	 }
      }
   }
}

void TelnetPort::Eof (void) {
   FUNCTION;
   Close();
   G->LOG_MISC.Put(G->Lang( 0x162, "ds", localport, strerror(errno)));
}

void TelnetPort::Display2 (Socket * where) {
   FUNCTION;
   where->Send(G->Lang( 0x163, "s", nick));
}

int TelnetPort::Dump ( char *, int ) {
   FUNCTION;
   return -1;
}

int TelnetPort::ExpMem ( void ) {
   FUNCTION;
   return sizeof(TelnetPort);
}

TelnetPort * TelnetPort::Find ( int port ) {
   FUNCTION;
   TelnetPort * ptr = (TelnetPort *)G->socklist;

   while ((ptr = ptr->Next()))
     if ((ptr->LocalPort() == port))
       break;
   return ptr;
}     
