/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "blowfish.h"
#include "ModuleEntryBlowfish.h"
#include "BlowfishBox.h"
#include "Socket.h"
#define CLASS "ModuleEntryBlowfish"

/* of course, if you change either of these, then your userfile will */
/* no longer be able to be shared. :)  */
#define SALT1  0xdeadd061
#define SALT2  0x23f6b095
static Version blowfish_ver;
static Version required_eggdrop;
static struct patches blowfish_patches[] = {
#include "patch.h"
};

ModuleEntryBlowfish::ModuleEntryBlowfish ( ModuleEntry * o, char ** r )
: ModuleEntryEncryption ( "blowfish", blowfish_ver.Set(2,0),
			 blowfish_patches, o, r) {
   FUNCTION;
   if (!*r) {
      if (!Depend("eggdrop", required_eggdrop.Set(2,0)))
	*r = "Your version of eggdrop is incompatible.";
      else {
	 G->module_encrypt = this;
	 memtype_blowfish = NEW ( MemType("blowfish", (List **)&boxlist, this) );
	 boxlist = NEW ( BlowfishBox () );
      }
   }
}
			   
ModuleEntryBlowfish::~ModuleEntryBlowfish ( void ) {
   FUNCTION;
   DELETE ( boxlist );
   if (G->module_encrypt == this)
      G->module_encrypt = 0;
}

int ModuleEntryBlowfish::ExpMem ( void ) {
   FUNCTION;
   return ModuleEntry::ExpMem () + sizeof(ModuleEntryBlowfish);
}

/* convert 64-bit encrypted password to text for userfile */
static char *base64 = "./0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";

static int base64dec (char c) {
   FUNCTION;
   int i;
   for (i = 0; i < 64; i++)
     if (base64[i] == c)
       return i;
   return 0;
}

void ModuleEntryBlowfish::EncryptPass (char * text, char * newp) {
   FUNCTION;
   UWORD_32bits left, right;
   int n;
   char *p;
   BlowfishBox * bb = boxlist->Find(text);
   
   if (!bb)
     bb = NEW ( BlowfishBox ( text, this ) );
   bb->Init();
   left = SALT1;
   right = SALT2;
   bb->Encipher(&left, &right);
   p = newp;
   *p++ = '+';			/* + means encrypted pass */
   for (n = 32; n > 0; n -= 6) {
      *p++ = base64[right & 0x3f];
      right = (right >> 6);
   }
   for (n = 32; n > 0; n -= 6) {
      *p++ = base64[left & 0x3f];
      left = (left >> 6);
   }
   *p = 0;
}

/* returned string must be freed when done with it! */
char * ModuleEntryBlowfish::Encrypt (const char * key, const char * str) {
   FUNCTION;
   UWORD_32bits left, right;
   char *p, *s, *dest, *d;
   int i;
   BlowfishBox * bb = boxlist->Find(key);

   if (!bb)
     bb = NEW ( BlowfishBox ( key, this ) );
   dest = (char *) my_malloc((strlen(str) + 9) * 2,G->memtype_misc);
   /* pad fake string with 8 bytes to make sure there's enough */
   s = (char *) my_malloc(strlen(str) + 9,memtype_blowfish);
   p = s + G->StrCpy(s, str);
   for (i = 0; i < 8; i++)
     *p++ = 0;
   bb->Init();
   p = s;
   d = dest;
   while (*p) {
      left = ((*p++) << 24);
      left += ((*p++) << 16);
      left += ((*p++) << 8);
      left += (*p++);
      right = ((*p++) << 24);
      right += ((*p++) << 16);
      right += ((*p++) << 8);
      right += (*p++);
      bb->Encipher(&left, &right);
      for (i = 0; i < 6; i++) {
	 *d++ = base64[right & 0x3f];
	 right = (right >> 6);
      }
      for (i = 0; i < 6; i++) {
	 *d++ = base64[left & 0x3f];
	 left = (left >> 6);
      }
   }
   *d = 0;
   my_free(s,memtype_blowfish);
   return dest;
}


/* returned string must be freed when done with it! */
char *ModuleEntryBlowfish::Decrypt (const char * key, const char * str) {
   FUNCTION;
   UWORD_32bits left, right;
   char *p, *s, *dest, *d;
   int i;
   BlowfishBox * bb = boxlist->Find(key);
   
   if (!bb)
     bb = NEW ( BlowfishBox ( key ) );
   dest = (char *) my_malloc(strlen(str) + 12,G->memtype_misc);
   /* pad encoded string with 0 bits in case it's bogus */
   s = (char *) my_malloc(strlen(str) + 12,memtype_blowfish);
   p = s + G->StrCpy(s, str);
   for (i = 0; i < 12; i++)
      *p++ = 0;
   bb->Init();
   p = s;
   d = dest;
   while (*p) {
      right = 0L;
      left = 0L;
      for (i = 0; i < 6; i++)
	 right |= (base64dec(*p++)) << (i * 6);
      for (i = 0; i < 6; i++)
	 left |= (base64dec(*p++)) << (i * 6);
      bb->Decipher(&left, &right);
      for (i = 0; i < 4; i++)
	 *d++ = (left & (0xff << ((3 - i) * 8))) >> ((3 - i) * 8);
      for (i = 0; i < 4; i++)
	 *d++ = (right & (0xff << ((3 - i) * 8))) >> ((3 - i) * 8);
   }
   *d = 0;
   my_free(s,memtype_blowfish);
   return dest;
}

void ModuleEntryBlowfish::Report (Socket * sock, int details) {
   FUNCTION;
   if (details) {
      BlowfishBox * bb = boxlist;

      sock->Printf("    Blowfish encryption module:\n");
      sock->Printf("    %d of %d boxes in use: ", boxlist->Count(), boxlist->Boxes());
      while ((bb = bb->Next()))
	sock->Printf("(age: %d) ", G->now - bb->lastuse);
      sock->Printf("\n");
   }
}
