/* 
 * FileDB.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "filesys.h"
#include "FileDB.h"
#include "UserEntryDccDir.h"
#include "UserEntryFstat.h"
#include "ModuleEntry.h"
#include "ModuleEntryFilesys.h"
#include "IntegerVariable.h"
#include "StringVariable.h"
#include "Socket.h"
#include "HelpReference.h"
#include <fcntl.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#if HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_NDIR_H
#include <ndir.h>
#endif
#endif
#define CLASS "FileDB"

FileDB * FILEDB;

FileDB::FileDB ( void ) : List () {
   FUNCTION;
}

FileDB::~FileDB ( void ) {
   FUNCTION;
}

FILE *FileDB::Open ( const char * path, int sort ) {
   FUNCTION;
   char s[DIRLEN+1], npath[DIRLEN+1];
   FILE *f;
   filedb fdb;
   struct stat st;
   struct flock fl;
   int l;
   
   if (count >= 2) 
     G->LOG_MISC.Put(FILESYS->Lang(0x002, "d", count));
   G->SPrintF(npath, DIRLEN, "%s%s", FILESYS->dccdir, safe_str(path));
   /* use alternate filename if requested */
   if (FILESYS->filedb_path[0]) {
      char s2[DIRLEN + 1], *p = strcpy(s2, safe_str(path));

      while (*p++)
	if (*p == '/')
	  *p = '.';
      l = G->SPrintF(s, DIRLEN, "%sfiledb.%s", FILESYS->filedb_path, s2);
      if (s[l - 1] == '.')
	 s[l - 1] = 0;
   } else
      G->SPrintF(s, DIRLEN, "%s/.filedb", npath);
   fl.l_type = F_WRLCK;
   fl.l_start = 0;
   fl.l_whence = SEEK_SET;
   fl.l_len = 0;   
   if (!(f = fopen(s, "r+b"))) {
      if ((f = fopen(s, "w+b"))) {
	 /* create new database and fix it up */
	 fcntl(fileno(f), F_SETLKW, &fl);
	 Update(npath, f, sort);
	 count++;
      }
   } else {
      /* lock it from other bots: */
      fcntl(fileno(f), F_SETLKW, &fl);
      /* check the timestamp... */
      fread(&fdb, sizeof(filedb), 1, f);
      ::stat(npath, &st);
      /* update filedb if: */
      /*  + it's been 6 hours since it was last updated */
      /*  + the directory has been visibly modified since then */
      /* (6 hours may be a bit often) */
      if (sort || ((G->now - fdb.timestamp) > (6 * 3600)) 
	  || (fdb.timestamp < st.st_mtime) ||
	  (fdb.timestamp < st.st_ctime)
	  || (fdb.version <= FILEVERSION_OLD)) 
	/* file database isn't up-to-date! */
	Update(npath, f, sort & 1);
      count++;
   }
   return f;
}

void FileDB::Close ( FILE * f ) {
   FUNCTION;
   struct flock fl;
   
   Timestamp(f);
   fseek(f, 0L, SEEK_END);
   count--;
   fl.l_type = F_UNLCK;
   fl.l_start = 0;
   fl.l_whence = SEEK_SET;
   fl.l_len = 0;
   fcntl(fileno(f), F_SETLKW, &fl);
   fclose(f);
}

void FileDB::Timestamp (FILE * f) {
   FUNCTION;
   filedb fdb;
   /* read 1st filedb entry if it's there */
   rewind(f);
   if (fread(&fdb, sizeof(filedb), 1, f) < 1) {
      /* MAKE 1st filedb entry then! */
      fdb.version = FILEVERSION;
      fdb.stat = FILE_UNUSED;
   }
   fdb.timestamp = G->now;
   rewind(f);
   fwrite(&fdb, sizeof(filedb), 1, f);
}

void FileDB::Update (char * path, FILE * f, int sort) {
   FUNCTION;
   struct dirent *dd;
   DIR *dir;
   filedb fdb[2];
   char name[61];
   long where, oldwhere;   
   struct stat st;
   char s[512];
   int ret;
   
   /* FIRST: make sure every real file is in the database */
   if (!(dir = opendir(path))) 
      G->LOG_MISC.Put(FILESYS->Lang(0x003, ""));
   else {
      for (dd = readdir(dir); dd; dd = readdir(dir)) {
	 if ((ret = NAMLEN(dd)) >= 60)
	   ret = 60;
	 strncpy(name, dd->d_name, ret);
	 name[ret] = 0;
	 if (ret != (int)NAMLEN(dd)) {
	    char s1[DIRLEN + 1], s2[DIRLEN + 1];
	    /* truncate name on disk */
	    if ((ret = G->SPrintF(s1, DIRLEN, "%s/", path)) + NAMLEN(dd)
		>= DIRLEN)
	      continue; /* just forget about this file */
	    strncpy(s1 + ret, dd->d_name, NAMLEN(dd));
	    s1[ret + NAMLEN(dd)] = 0;
	    G->SPrintF(s2, DIRLEN, "%s/%s", path, name);
	    G->MoveFile(s1, s2);
	 }
	 if (name[0] != '.') {
	    G->SPrintF(s, DIRLEN, "%s/%s", path, name);
	    ::stat(s, &st);
	    where = 0;
	    if (!(ret = FILESYS->FindMatch(f, name, &where, &fdb[0]))) {
	       /* new file! */
	       where = FILESYS->FindEmpty(f);
	       G->memset(&fdb,0,sizeof(filedb));
	       fdb[0].version = FILEVERSION;
	       strcpy(fdb[0].filename, name);
	       strcpy(fdb[0].uploader, G->botnetnick);
	       fdb[0].uploaded = G->now;
	       if (S_ISDIR(st.st_mode))
		 fdb[0].stat |= FILE_DIR;
	    } else if (FILEVERSION >= fdb[0].version) {
	       /* old version filedb, do the dirty */
	       if (fdb[0].version == FILEVERSION_OLD) {
		  filedb_old * fdbo = (filedb_old *)&fdb[0];
		  fdb[0].desc[185] = 0; /* truncate it */
		  fdb[0].chname[0] = 0; /* new entry */
		  strcpy(fdb[0].uploader,fdbo->uploader); /* moved forward a few bytes*/
		  strcpy(fdb[0].flags_req,fdbo->flags_req);/* and again */
		  fdb[0].version = FILEVERSION;
	       } else {
		  G->LOG_MISC.Put(FILESYS->Lang(0x001, ""));
		  continue;
	       }
	    }
	    /* update size if needed */
	    fdb[0].size = st.st_size;
	    fseek(f, where, SEEK_SET);
	    fwrite(&fdb[0], sizeof(filedb), 1, f);
	 }
      }
      closedir(dir);
      /* SECOND: make sure every db file is real, and sort as we go, if we're
       * sorting */
      rewind(f);
      while (!feof(f)) {
	 where = ftell(f);
	 fread(&fdb[0], sizeof(filedb), 1, f);
	 if (!feof(f)) {
	    if (!(fdb[0].stat & FILE_UNUSED) && !fdb[0].sharelink[0]) {
	       G->SPrintF(s, DIRLEN, "%s/%s", path, fdb[0].filename);
	       if (::stat(s, &st) != 0) {
		  /* gone file */
		  fseek(f, where, SEEK_SET);
		  fdb[0].stat |= FILE_UNUSED; 
		  fwrite(&fdb[0], sizeof(filedb), 1, f);
		  /* sunos and others will puke bloody chunks if you write the */
		  /* last record in a file and then attempt to read to EOF: */
		  fseek(f, where, SEEK_SET);
		  continue; /* cycle to next one */
	       }
	    }
	    if (sort && !(fdb[0].stat & FILE_UNUSED)) {
	       rewind(f);
	       oldwhere = ftell(f);
	       ret = 0;
	       while (!feof(f) && (oldwhere < where)) {
		  fread(&fdb[1 - ret], sizeof(filedb), 1, f);
		  if (!feof(f)) {
		     if ((fdb[0].stat & FILE_UNUSED) ||
			 (strcasecmp(fdb[ret].filename,fdb[1-ret].filename) < 0)) {
			/* our current is < the checked one, insert here */
			fseek(f, oldwhere, SEEK_SET);
			fwrite(&fdb[ret], sizeof(filedb), 1, f);
			ret = 1 - ret;
			/* and fall out */
		     }
		     /* otherwise read next entry */
		     oldwhere = ftell(f);
		  }
	       }
	       /* here, either we've found a place to insert, or got to
		* the end of the list */
	       /* if we've got to the end of the current list oldwhere == where
		* so we fall through */
	       /* ret will point to the current valid one */
	       while (!feof(f) && (oldwhere < where)) {
		  /* need to move this entry up 1 .. */
		  fread(&fdb[1-ret], sizeof(filedb), 1, f);
		  oldwhere = ftell(f);
		  /* write lower record here */
		  fseek(f, oldwhere, SEEK_SET);
		  fwrite(&fdb[ret], sizeof(filedb), 1, f);
		  ret = 1 - ret;
	       }
	       /* when we get here fdb[ret] holds the last record,
		* which needs to be written where we first grabbed the record
		* from
		*/
	       fseek(f,where,SEEK_SET);
	       fwrite(&fdb[ret], sizeof(filedb),1,f);
	    }  
	 } 
      }
      /* write new timestamp */
      Timestamp(f);
   }
}
