/* 
 * ModuleEntryFilesys.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "filesys.h"
#include "UserEntryDccDir.h"
#include "UserEntryFstat.h"
#include "ModuleEntryFilesys.h"
#include "IntegerVariable.h"
#include "StringVariable.h"
#include "Socket.h"
#include "HelpReference.h"
#include "FileDB.h"
#include <fcntl.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#if HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_NDIR_H
#include <ndir.h>
#endif
#endif
#define CLASS "ModuleEntryFilesys"

static Version filesys_version;
static struct patches filesys_patches[] = { 
#include "patch.h"
};

ModuleEntryFilesys::ModuleEntryFilesys( ModuleEntry * old, char ** r) 
: ModuleEntry ( "filesys", filesys_version.Set(2,0), filesys_patches, old ) {
   FUNCTION;
   Version required;
   
   FILESYS = this;
   if (!Depend("eggdrop", required.Set(2,0)))
     *r = "Your eggdrop core is incompatible.";
   else if (!(BOTNET = (ModuleEntryBotnet *) Depend("botnet", required.Set(2,0))))
     *r = "You need the botnet module to use the filesys module.";
   else if (!(SERVER = (ModuleEntryServer *) Depend("server", required.Set(2,0))))
     *r = "You need the server module to use the filesys module.";
   else if (!(TRANSFER = (ModuleEntryTransfer *) Depend("transfer", required.Set(2,0))))
     *r = "You need the transfer module to use the filesys module.";
   else {
      dcc_maxsize = 1024;
      dcc_users = 0;
      upload_to_cd = 0;
      dccdir[0] = 0;
      dccin[0] = 0;
      filedb_path[0] = 0;
      NEW ( HelpReference("filesys.help",this) );
      NEW ( IntegerVariable ( "max_filesize", &dcc_maxsize, 0, this ) );
      NEW ( IntegerVariable ( "max_file_users", &dcc_users, 0, this ) );
      NEW ( IntegerVariable ( "upload_to_pwd", &upload_to_cd, 0, this ) );
      NEW ( StringVariable ( "files_path", dccdir, 120, 0, this ) );
      NEW ( StringVariable ( "incoming_path", dccin, 120, 0, this ) );
      NEW ( StringVariable ( "filedb_path", filedb_path, 120, 0, this ) );
      entrytype_dccdir = NEW ( UserEntryDccDir (this) );
      entrytype_fstat = NEW ( UserEntryFstat (this) );
      FILEDB = NEW ( FileDB );
   }
}

ModuleEntryFilesys::~ModuleEntryFilesys ( void ) {
   FUNCTION;
   Socket * ptr = G->socklist, *old = ptr;
   
   context;
   DELETE ( FILEDB );
   G->LOG_MISC.Put(FILESYS->Lang(0x0CF, ""));
   while ((ptr = old->Next())) {
      if (ptr->Type() == FILESYS->DCC_FILES) {
	 ptr->Send(FILESYS->Lang(0x103, ""));
	 ptr->Send(FILESYS->Lang(0x0D0, ""));
	 delete ptr;
      } else if (ptr->Type() == FILESYS->DCC_FILES_PASS)
	   delete ptr;
      else 
	old = ptr;
   }
   context;
}

int ModuleEntryFilesys::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ModuleEntryFilesys) + ModuleEntry::ExpMem ();
}

void ModuleEntryFilesys::Report ( Socket * idx, int ) {
   FUNCTION;
   if (FILESYS->dccdir[0]) {
      idx->Send(FILESYS->Lang(0x0C8, "s", FILESYS->dccdir));
      idx->Send(FILESYS->Lang(FILESYS->upload_to_cd ? 0x0C9 : (FILESYS->dccin[0] ? 0x0CA : 0x0CB),
			   "s" , FILESYS->dccin));
      if (FILESYS->dcc_users)
	idx->Send(FILESYS->Lang(0x0CC, "d", FILESYS->dcc_users));
      if (FILESYS->upload_to_cd || FILESYS->dccin[0])
	idx->Send(FILESYS->Lang(0x0CD, "d", FILESYS->dcc_maxsize));
   } else
     idx->Send(FILESYS->Lang(0x0CE, ""));
}

/* are there too many people in the file system? */
int ModuleEntryFilesys::TooManyFilers ( void ) {
   FUNCTION;
   int n = 0;
   Socket * ptr = G->socklist;
   
   if (!FILESYS->dcc_users)
     return 0;
   while ((ptr = ptr->Next()))
     if (ptr->Type() == FILESYS->DCC_FILES)
       n++;
   return (n >= FILESYS->dcc_users);
}

/* use a where of 0 to start out, then increment 1 space for each next */
int ModuleEntryFilesys::FindMatch (FILE * f, const char * lookfor, long * where, filedb * fdb) {
   FUNCTION;
   char match[256];
   int l = strlen(lookfor);
   
   if (l > 255)
     l = 255;
   if (lookfor[l - 1] == '/')
     l --;
   strncpy(match, lookfor, l); 
   match[l] = 0;
   fseek(f, *where, SEEK_SET);
   while (!feof(f)) {
      *where = ftell(f);
      fread(fdb, sizeof(filedb), 1, f);
      if (!feof(f) && !(fdb->stat & FILE_UNUSED) 
	  && FILESYS->TRANSFER->WildMatchFile((unsigned char *)match, 
				       (unsigned char *)fdb->filename))
	return 1;
   }
   return 0;
}

long ModuleEntryFilesys::FindEmpty (FILE * f) {
   FUNCTION;
   long where = 0L;
   filedb fdb;
   
   rewind(f);
   while (!feof(f)) {
      where = ftell(f);
      fread(&fdb, sizeof(filedb), 1, f);
      if (!feof(f) && (fdb.stat & FILE_UNUSED))
	return where;
   }
   fseek(f, 0L, SEEK_END);
   return ftell(f);
}

int FileDB::ExpMem ( void ) {
   FUNCTION;
   return 0;
}
