/*
 * TclScript.C - code to support TCL Scripting
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "TclScript.h"
#include "ScriptArgument.h"
#include "FlagRecord.h"
#include "BindEntry.h"
#include "IntegerVariable.h"
#include "CoupletVariable.h"
#include "StringVariable.h"
#include "MiscVariable.h"
#include "ModuleEntryTcl.h"
#include "Command.h"
#define CLASS "TclScript"

/* read/write integer couplets (int1:int2) */
static char *tcl_eggcouplet (ClientData cdata, Tcl_Interp * interp, char * name1,
			    char * name2, int flags) {
   FUNCTION;
   char *s, s1[41];
   CoupletVariable * cp = (CoupletVariable *)cdata;
   
   if (flags & (TCL_TRACE_READS | TCL_TRACE_UNSETS)) {
      T->G->SPrintF(s1, 40, "%d:%d", *(cp->lptr), *(cp->rptr));
      Tcl_SetVar2(interp, name1, name2, s1, TCL_GLOBAL_ONLY);
   } else if ((s = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY))) {
      T->G->SplitC(s1, s, ':');
      if (s1[0]) {
	 *(cp->lptr) = atoi(s1);
	 *(cp->rptr) = atoi(s);
      } else
	*(cp->lptr) = atoi(s);
   }
   return NULL;
}

/* read/write normal integer */
static char *tcl_eggint (ClientData cdata, Tcl_Interp * irp,
			 char * name1, char * name2, int flags) {
   FUNCTION;
   char *s, s1[41];
   long l;
   IntegerVariable * iv = (IntegerVariable *)cdata;
   
   if (flags & (TCL_TRACE_READS | TCL_TRACE_UNSETS)) {
      /* special cases */
      if (iv->val == (int *)&T->G->conmask)
	 strcpy(s1, T->G->conmask.Mask());
      else if (iv->val == &T->G->default_flags) {
	 FlagRecord fr;
	 fr.SetGlobal(T->G->default_flags);
	 fr.Build(s1);
      } else
	T->G->SPrintF(s1, 40, "%d", *(iv->val));
      Tcl_SetVar2(irp, name1, name2, s1, TCL_GLOBAL_ONLY);
   } else if ((s = Tcl_GetVar2(irp, name1, name2, TCL_GLOBAL_ONLY))) {
      if (iv->val == (int *)&T->G->conmask) {
	 if (s[0])
	   T->G->conmask.Set(s);
	 else
	   T->G->conmask.Set ( LogModes::MODES | LogModes::MISC | LogModes::CMDS);
      } else if (iv->val == &T->G->default_flags) {
	 FlagRecord fr;
	 fr.BreakDown(s);
	 T->G->default_flags = fr.Global();
      } else if ((iv->readonly == 2) 
		 || ((iv->readonly == 1) && T->G->protect_readonly))
	   return T->G->Lang(0x129, "");
      else {
	 if (Tcl_ExprLong(irp, s, &l) == TCL_ERROR)
	   return irp->result;
	 *(iv->val) = (int) l;
      }
   }
   return NULL;
}

/* read/write normal string variable */
static char *tcl_eggstr (ClientData cdata, Tcl_Interp * interp,
			 char * name1, char * name2, int flags) {
   FUNCTION;
   char *s;
   StringVariable *st = (StringVariable *) cdata;
   
   if (flags & (TCL_TRACE_READS | TCL_TRACE_UNSETS)) {
      Tcl_SetVar2(interp, name1, name2, st->buf, TCL_GLOBAL_ONLY);
      if (st->flags & (VAR_PROTECT | VAR_READONLY)) {
	 if ((flags & TCL_TRACE_UNSETS) && 
	     (T->G->protect_readonly || (st->flags & VAR_READONLY))) {
	    /* no matter what we do, it won't return the error */
	    Tcl_SetVar2(interp, name1, name2, st->buf, TCL_GLOBAL_ONLY);
	    Tcl_TraceVar(interp, name1, TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
			 tcl_eggstr, (ClientData) st);
	    return T->G->Lang(0x129, "");
	 }
      }
   } else {			/* writes */
      if ((st->flags & (VAR_PROTECT | VAR_READONLY))
	   && (T->G->protect_readonly || (st->flags & VAR_READONLY))) {
	 Tcl_SetVar2(interp, name1, name2, st->buf, TCL_GLOBAL_ONLY);
	 return T->G->Lang(0x129, "");
      }
      if ((s = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY))) {
	 strncpy(st->buf, s, st->length);
	 st->buf[st->length] = 0;
	 if ((st->flags & VAR_DIR) && s[0]) {
	    int l = strlen(st->buf) - 1;
	    
	    if (st->buf[l] != '/') {
	       st->buf[l++] = '/';
	       st->buf[l] = 0;
	    }
	 }
      }
   }
   return NULL;
}

static char *tcl_eggmisc (ClientData cdata, Tcl_Interp * interp,
			  char * name1, char * name2, int flags) {
   FUNCTION;
   MiscVariable * mv = (MiscVariable *)cdata;
   char * res, * old = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
   
   if (flags & TCL_TRACE_READS) 
     Tcl_SetVar2(interp, name1, name2, (res = mv->Read(T->entry, 0, 0)), TCL_GLOBAL_ONLY);
   else if (flags & TCL_TRACE_UNSETS) {
      if ((res = mv->Clear(T->entry, 0, 0))) 
	Tcl_SetVar2(interp, name1, name2, res, TCL_GLOBAL_ONLY);
   } else {
      char * s = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
      if (mv->flags & VAR_LIST) {
	 int lc, lc2;
	 char ** list;
	 ScriptArgument ** args;
	 
	 if (Tcl_SplitList(interp, s, &lc, &list) != TCL_ERROR) {
	    args = (ScriptArgument **)
	      my_malloc(sizeof(ScriptArgument *) * (lc + 1),T->G->memtype_misc);
	    args[lc2 = lc] = 0;
	    while (lc) {
	       lc--;
	       args[lc] = NEW ( ScriptArgument ( T->entry, list[lc] ) );
	    }
	    if ((res = mv->Set(T->entry, lc2, args)))  {
	       Tcl_AppendResult(interp, res, NULL);
	       Tcl_SetVar2(interp, name1, name2, old, TCL_GLOBAL_ONLY);
	    }
	    while (lc2) {
	       lc2--;
	       DELETE ( args[lc2] );
	    }
	    DELETE (args);
	 } else
	   return interp->result;
      } else {
	 ScriptArgument pass(T->entry, s), * args[2] = {&pass, 0};
	 
	 if ((res = mv->Set(T->entry, 1, args)))  {
	    Tcl_AppendResult(interp, res, NULL);
	    Tcl_SetVar2(interp, name1, name2, old, TCL_GLOBAL_ONLY);
	 }
      }
   }
   if (res && (mv->flags & VAR_LIST))
     my_free ( res, T->G->memtype_misc );
   return NULL;
}

static int tcl_func_callback ( ClientData cd, Tcl_Interp *, int argc, char ** argv) {
   FUNCTION;
   ScriptArgument * args[256];
   int i;
   Command * ci = (Command *)cd;
   
   if (argc > 255)
     argc = 255;
   for (i = 0; i < argc; i++) 
     args[i] = NEW ( ScriptArgument ( T->entry, argv[i] ) );
   args[i] = 0;
   if (ci->Call(T->entry, i, args) == SCRIPT_OK)
     return TCL_OK;
   return TCL_ERROR;
}
   
TclScript::~TclScript ( void ) {
   FUNCTION;
   Tcl_DeleteInterp(interp);
}

int TclScript::ExpMem ( void ) { 
   FUNCTION;
   return sizeof(TclScript);
}

const char * TclScript::AsString ( void ) {
   FUNCTION;
   return interp->result;
}

int TclScript::AsInteger ( void ) {
   FUNCTION;
   return atoi(interp->result);
}

void TclScript::Clear ( void ) {
   FUNCTION;
   Tcl_ResetResult(interp);
}

void TclScript::SetString ( const char * st ) {
   FUNCTION;
   Tcl_SetResult(interp, (char *)st, TCL_VOLATILE);
}

void TclScript::SetInteger ( int x ) {
   FUNCTION;
   Tcl_SetResult(interp, G->ToBase10(x), TCL_VOLATILE);
}

void TclScript::SetArray ( int s, const char ** list) {
   FUNCTION;
   char * p = Tcl_Merge(s, (char **)list);
   Tcl_SetResult(interp, p, TCL_VOLATILE);
   Tcl_Free(p);
}

void TclScript::AppendString ( const char * st ) {
   FUNCTION;
   Tcl_AppendResult(interp, st, NULL );
}

void TclScript::AppendInteger ( int x ) {
   FUNCTION;
   Tcl_AppendResult(interp, G->ToBase10(x), NULL );
}

void TclScript::AppendArray ( int s, const char ** list) {
   FUNCTION;
   char * p = Tcl_Merge(s, (char **)list);
   Tcl_AppendElement(interp, p);
   Tcl_Free(p);
}

void TclScript::AddStringVar ( StringVariable * sv ) {
   FUNCTION;
   if (sv->length > 0) {
      char *p = Tcl_GetVar(interp, (char *)sv->Key(), TCL_GLOBAL_ONLY);
      if (p) {
	 strncpy(sv->buf, p, sv->length);
	 sv->buf[sv->length] = 0;
	 if (sv->flags & VAR_DIR) {
	    int x = strlen(sv->buf);
	    
	    if ((x > 0) && (x < (sv->length - 1))
		&& (sv->buf[x - 1] != '/')) {
	       sv->buf[x++] = '/';
	       sv->buf[x] = 0;
	    }
	 }
      }
   }
   if (Tcl_TraceVar(interp, (char *)sv->Key(), TCL_TRACE_READS | TCL_TRACE_WRITES |
		    TCL_TRACE_UNSETS, tcl_eggstr, (ClientData) sv) == TCL_ERROR)
     G->LOG_MISC.Put(T->Lang(0x005, "s", interp->result));
}
   
void TclScript::RemStringVar ( StringVariable * sv ) {
   FUNCTION;
   if (sv == (StringVariable *)Tcl_VarTraceInfo(interp, (char *)sv->Key(), TCL_TRACE_READS 
					   | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
					   tcl_eggstr, NULL))
     Tcl_UntraceVar(interp, (char *)sv->Key(),
		    TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		    tcl_eggstr, sv);
}

void TclScript::AddIntVar ( IntegerVariable * iv ) {
   FUNCTION;
   char *p;
   
   if ((iv->readonly != VAR_READONLY) 
       && (p = Tcl_GetVar(interp, (char *)iv->Key(), TCL_GLOBAL_ONLY)))
     *(iv->val) = atoi(p);
   if (Tcl_TraceVar(interp, (char *)iv->Key(),
		    TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		    tcl_eggint, (ClientData) iv) == TCL_ERROR)
     G->LOG_MISC.Put(T->Lang(0x006, "s", interp->result));
}

void TclScript::RemIntVar ( IntegerVariable * iv) {
   FUNCTION;
   if (iv == (IntegerVariable *) Tcl_VarTraceInfo(interp, (char *)iv->Key(),
					 TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
					 tcl_eggint, NULL)) {
      Tcl_UntraceVar(interp, (char *)iv->Key(),
		     TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		     tcl_eggint, (ClientData) iv);
   }
}

void TclScript::AddCoupletVar ( CoupletVariable * cv ) {
   FUNCTION;
   char *p = Tcl_GetVar(interp, (char *)cv->Key(), TCL_GLOBAL_ONLY);
   if (p)
     tcl_eggcouplet ((ClientData)cv, interp, (char *)cv->Key(), 0, 0);
   if (Tcl_TraceVar(interp, (char *)cv->Key(),
		    TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		    tcl_eggcouplet, (ClientData) cv) == TCL_ERROR)
   G->LOG_MISC.Put(T->Lang(0x007, "s", interp->result));
}

void TclScript::RemCoupletVar ( CoupletVariable * cv ) {
   FUNCTION;
   if (cv == (CoupletVariable *) Tcl_VarTraceInfo(interp, (char *)cv->Key(),
					     TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
					     tcl_eggcouplet, NULL)) {
      Tcl_UntraceVar(interp, (char *)cv->Key(),
		     TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		     tcl_eggcouplet, (ClientData) cv);
   }
}

void TclScript::AddMiscVar ( MiscVariable * mv ) {
   FUNCTION;
   char *p = Tcl_GetVar(interp, (char *)mv->Key(), TCL_GLOBAL_ONLY);
   if (p)
     tcl_eggmisc ((ClientData)mv, interp, (char *)mv->Key(), 0, 0);
   if (Tcl_TraceVar(interp, (char *)mv->Key(),
		    TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		    tcl_eggmisc, (ClientData) mv) == TCL_ERROR)
     G->LOG_MISC.Put(T->Lang(0x008, "s", interp->result));
}

void TclScript::RemMiscVar ( MiscVariable * mv) {
   FUNCTION;
   if (mv == (MiscVariable *) Tcl_VarTraceInfo(interp, (char *)mv->Key(),
					  TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
					  tcl_eggmisc, NULL)) {
      Tcl_UntraceVar(interp, (char *)mv->Key(),
		     TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS,
		     tcl_eggmisc, (ClientData) mv);
   }
}

void TclScript::AddCommand ( Command * ci ) {
   FUNCTION;
   Tcl_CreateCommand(interp, (char *)ci->Key(), tcl_func_callback, 
		     (ClientData)ci, 0);
}

void TclScript::RemCommand ( Command * ci ) {
   FUNCTION;
   Tcl_DeleteCommand(interp, (char *)ci->Key());
}

/* trigger (execute) a proc */
BindReturn TclScript::CallBind (BindEntry * be, const char * bind, int argc, 
				va_list valist) {
   FUNCTION;
   static int argcntr = 0;
   FILE * f = 0;
   int l;
   char buf[512], * p = buf;

   while (argc) {
      l = G->SPrintF(p, 511, " $_a%d", argcntr++);
      Tcl_SetVar(interp, p + 2, va_arg(valist, char *), TCL_GLOBAL_ONLY);
      p += l;
      argc--;
   }
   if (T->debug_tcl && !(f = fopen("DEBUG.TCL", "a")))
     fprintf(f, "eval: %s%s\n", be->Key(), buf);
   Tcl_SetVar(interp, "tcl_interactive", "0", TCL_GLOBAL_ONLY);
   context;
   Tcl_SetVar(interp,"lastbind", (char *)bind, TCL_GLOBAL_ONLY);
   if (Tcl_VarEval(interp, (char *)be->Key(), buf, NULL) == TCL_ERROR) {
      if (T->debug_tcl && f) {
	 fprintf(f, "done eval. error.\n");
	 fclose(f);
      }
      G->LOG_MISC.Put(T->Lang(0x004, "ss", be->Key(), interp->result));
      return EXECUTED;
   } else {
      if (T->debug_tcl && f) {
	 fprintf(f, "done eval. ok.\n");
	 fclose(f);
      }
      if (!strcmp(interp->result, "break"))
	return EXEC_BRK;
      return (atoi(interp->result) > 0) ? EXEC_LOG : EXECUTED;
   }
}

int TclScript::Evaluate ( const char * script, int argc, ... ) {
   FUNCTION;
   int code, l = 0;
   FILE * f = 0;
   char args[512];
   va_list valist;
   
   va_start(valist, argc);
   args[0] = 0;
   for (code = 0; (code < argc) && (l < 511); code++)
     G->SPrintF(args + l, 511 - l, " %s", va_arg(valist, char *));
   va_end(valist);
   if (T->debug_tcl && ((f = fopen("DEBUG->TCL", "a"))))
     fprintf(f, "eval: %s %s\n", script, args);
   Tcl_SetVar(interp, "tcl_interactive", "0", TCL_GLOBAL_ONLY);
   context;
   code = Tcl_VarEval(interp, script, args, 0);
   if (T->debug_tcl && f) {
      fprintf(f, "done eval, result=%d\n", code);
      fclose(f);
   }
   if (code != TCL_OK) {
      G->LOG_MISC.Put(G->Lang(0x12B, "ss", script, args));
      G->LOG_MISC.Put(interp->result);
   }
   return (code == TCL_OK) ? SCRIPT_OK : SCRIPT_ERROR;
}
